"""grouch.script_util

Code common to the Grouch scripts.
"""

# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: script_util.py,v 1.3 2001/09/05 14:21:23 gward Exp $"

import sys

def writenow (msg):
    sys.stdout.write(msg)
    sys.stdout.flush()

def announce (msg, threshold=1):
    import __main__
    if __main__.VERBOSITY >= threshold:
        writenow(msg)

def error (msg):
    sys.stderr.write("error: " + msg + "\n")

def warn (msg):
    sys.stderr.write("warning: " + msg + "\n")
