"""grouch

An system for describing and enforcing a Python object schema.
"""

# created 2000/12/13, Greg Ward
#
# Copyright (c) 2001 CNRI.  All Rights Reserved.

__revision__ = "$Id: __init__.py,v 1.3 2001/09/05 14:21:23 gward Exp $"
